@@ -1,5 +1,7 @@
+"""Classes and functions which implement HMM algorithms."""
+
 from functools import reduce
-from itertools import accumulate
+from itertools import chain, accumulate
 
 import scipy.stats as stats
 from numpy import exp, log
@@ -60,9 +62,10 @@ class HMM:
         if set(e_dists) != set(states):
             raise ValueError('States in e_dists do not match those in t_dists.')
         all_dicts = all([isinstance(e_dist, dict) for e_dist in e_dists.values()])
-        all_rvs = all([getattr(e_dist, 'pmf', None) and getattr(e_dist, 'rvs', None) for e_dist in e_dists.values()])
-        if not all_dicts and not all_rvs:
-            raise ValueError('e_dists are not either all dicts or all rvs.')
+        all_pmfs = all([getattr(e_dist, 'pmf', None) and getattr(e_dist, 'rvs', None) for e_dist in e_dists.values()])
+        all_pdfs = all([getattr(e_dist, 'pdf', None) and getattr(e_dist, 'rvs', None) for e_dist in e_dists.values()])
+        if not (all_dicts or all_pmfs or all_pdfs):
+            raise ValueError('e_dists are not either all dicts or all rvs of the same type.')
 
         # Check start_dist
         if not set(start_dist) <= states:
@@ -71,7 +74,7 @@ class HMM:
         # Create random variates from t_dists
         state2idx = {}
         idx2state = {}
-        for idx, state in enumerate(states):
+        for idx, state in enumerate(sorted(states)):
             state2idx[state] = idx
             idx2state[idx] = state
         t_dists_rv = {}
@@ -79,13 +82,13 @@ class HMM:
             idx = state2idx[state]
             t_dists_rv[idx] = rv_from_dict(t_dist, state2idx)
 
-        # Create random variates from e_dists
+        # Create random variates from e_dists if dictionaries
         if all_dicts:
             emits = {emit for e_dist in e_dists.values() for emit in e_dist}
             emit2idx = {}
             idx2emit = {}
             _emits = set()
-            for idx, emit in enumerate(emits):
+            for idx, emit in enumerate(sorted(emits)):
                 emit2idx[emit] = idx
                 idx2emit[idx] = emit
                 _emits.add(idx)
@@ -103,6 +106,10 @@ class HMM:
         # Create random variate from start_dist
         start_dist_rv = rv_from_dict(start_dist, state2idx)
 
+        # Extract probability functions from e_dists
+        attr = 'pdf' if all_pdfs else 'pmf'
+        e_dists_pf = {state: getattr(e_dist, attr) for state, e_dist in e_dists_rv.items()}
+
         self.name = name
         self.states = states
         self._states = set(idx2state)
@@ -115,6 +122,7 @@ class HMM:
         self._t_dists_rv = t_dists_rv
         self.e_dists = e_dists
         self._e_dists_rv = e_dists_rv
+        self._e_dists_pf = e_dists_pf
         self.start_dist = start_dist
         self._start_dist_rv = start_dist_rv
         self.stop_states = stop_states
@@ -142,16 +150,16 @@ class HMM:
         """
         if step_max == 0:
             return []
-        s0 = self._start_dist_rv.rvs()
-        e0 = self._e_dists_rv[s0].rvs()
-        steps = [(self._idx2state[s0], self._idx2emit[e0])]
+        state0 = self._start_dist_rv.rvs()
+        emit0 = self._e_dists_rv[state0].rvs()
+        steps = [(self._idx2state[state0], self._idx2emit[emit0])]
         for i in range(step_max-1):
-            if s0 in self._stop_states:
+            if state0 in self._stop_states:
                 return steps
-            s1 = self._t_dists_rv[s0].rvs()
-            e1 = self._e_dists_rv[s1].rvs()
-            steps.append((self._idx2state[s1], self._idx2emit[e1]))
-            s0 = s1
+            state1 = self._t_dists_rv[state0].rvs()
+            emit1 = self._e_dists_rv[state1].rvs()
+            steps.append((self._idx2state[state1], self._idx2emit[emit1]))
+            state0 = state1
         return steps
 
     def viterbi(self, emits):
@@ -171,18 +179,17 @@ class HMM:
         """
         # Forward pass
         emits = [self._emit2idx[emit] for emit in emits]  # Convert emits to internal labels
-        vs = {state: [(log(self._e_dists_rv[state].pmf(emits[0])) + log(self._start_dist_rv.pmf(state)), [None])]
-              for state in self._states}
+        vs = {state: [(log(self._e_dists_pf[state](emits[0])) + log(self._start_dist_rv.pmf(state)), [None])] for state in self._states}
         for i, emit in enumerate(emits[1:]):
-            for state in self._states:
+            for state0 in self._states:
                 # Get probabilities
-                t_probs = {s: vs[s][i][0] + log(self._t_dists_rv[s].pmf(state)) for s in self._states}
+                t_probs = {state1: vs[state1][i][0] + log(self._t_dists_rv[state1].pmf(state0)) for state1 in self._states}
                 t_prob = max(t_probs.values())  # Probability of most likely path to state
-                e_prob = log(self._e_dists_rv[state].pmf(emit))
+                e_prob = log(self._e_dists_pf[state0](emit))
 
                 # Get traceback states
                 tb_states = [s for s, p in t_probs.items() if p == t_prob]
-                vs[state].append((e_prob+t_prob, tb_states))
+                vs[state0].append((e_prob+t_prob, tb_states))
 
         # Compile traceback states (taking care to allow for multiple paths)
         v_max = max([v[-1][0] for v in vs.values()])
@@ -223,7 +230,7 @@ class HMM:
 
         # Initialize
         emits = [self._emit2idx[emit] for emit in emits]  # Convert emits to internal labels
-        fs = {state: [self._e_dists_rv[state].pmf(emits[0]) * self._start_dist_rv.pmf(state)] for state in self._states}
+        fs = {state: [self._e_dists_pf[state](emits[0]) * self._start_dist_rv.pmf(state)] for state in self._states}
         s = sum([fs[state][0] for state in self._states])
         for state in self._states:
             fs[state][0] /= s
@@ -232,11 +239,11 @@ class HMM:
         # Forward pass
         for i, emit in enumerate(emits[1:]):
             # Get probabilities
-            for state in self._states:
-                t_probs = [fs[s][i] * self._t_dists_rv[s].pmf(state) for s in self._states]
+            for state0 in self._states:
+                t_probs = [fs[state1][i] * self._t_dists_rv[state1].pmf(state0) for state1 in self._states]
                 t_prob = sum(t_probs)  # Probability of all paths to state
-                e_prob = self._e_dists_rv[state].pmf(emit)
-                fs[state].append(e_prob*t_prob)
+                e_prob = self._e_dists_pf[state0](emit)
+                fs[state0].append(e_prob*t_prob)
 
             # Scale probabilities
             s = sum([fs[state][i+1] for state in self._states])
@@ -281,10 +288,10 @@ class HMM:
         # Backward pass
         for i, emit in enumerate(emits[:0:-1]):  # Reverse sequence starting from last emit excluding first
             # Get probabilities
-            for state in self._states:
-                probs = [bs[s][i] * self._t_dists_rv[state].pmf(s) * self._e_dists_rv[s].pmf(emit) for s in self._states]
+            for state0 in self._states:
+                probs = [bs[state1][i] * self._t_dists_rv[state0].pmf(state1) * self._e_dists_pf[state1](emit) for state1 in self._states]
                 prob = sum(probs)  # Probability of all paths to state
-                bs[state].append(prob)
+                bs[state0].append(prob)
 
             # Scale probabilities
             s = sum([bs[state][i+1] for state in self._states])
@@ -297,6 +304,11 @@ class HMM:
     def forward_backward(self, emits):
         """Infer state probabilities at each time point.
 
+        The probabilities are not re-normalized, so the sums of posterior
+        probabilities across states at a given time point will not necessarily
+        equal one. This is done to not give false confidence in the numerical
+        precision of the results.
+
         Parameters
         ----------
         emits: list
@@ -304,9 +316,8 @@ class HMM:
 
         Returns
         -------
-        fb: list of dicts
-            List where each element is a dict of state probabilities keyed by
-            the state labels.
+        fb: dict of lists
+            State probabilities at each time point keyed by the state labels.
         """
         fs, ss_f = self.forward(emits)
         bs, ss_b = self.backward(emits)
@@ -323,6 +334,22 @@ class HMM:
 
 
 class ARHMM:
+    """A class for storing HMM model parameters and computing inferences from sequences of emissions.
+
+    The implementation is essentially identical to the HMM class with a few
+    exceptions to accommodate the emission distribution's dependence on the
+    previous emission. First, the pmf or pdf method of the emission
+    distributions must accept a tuple of the previous and current emissions,
+    respectively. (Consequently, dictionary style emission distributions are
+    not supported for ARHMMs.) The rvs method must also accept the previous
+    emission as its first argument. Second, an ARHMM requires additional start
+    emission distributions for each possible start state, as the behavior of
+    the emission distribution is technically undefined when there is no previous
+    emission.
+
+    Otherwise the ARHMM class is identical to the HMM class, so please refer to
+    it for addtional details.
+    """
     def __init__(self, t_dists, e_dists, start_t_dist, start_e_dists, stop_states=None, name='arhmm'):
         # Check t_dists
         states = set(t_dists)
@@ -333,9 +360,10 @@ class ARHMM:
         # Check e_dists
         if set(e_dists) != set(states):
             raise ValueError('States in e_dists do not match those in t_dists.')
-        all_rvs = all([getattr(e_dist, 'pmf', None) and getattr(e_dist, 'rvs', None) for e_dist in e_dists.values()])
-        if not all_rvs:
-            raise ValueError('e_dists are not all rvs.')
+        all_pmfs = all([getattr(e_dist, 'pmf', None) and getattr(e_dist, 'rvs', None) for e_dist in chain(e_dists.values(), start_e_dists.values())])
+        all_pdfs = all([getattr(e_dist, 'pdf', None) and getattr(e_dist, 'rvs', None) for e_dist in chain(e_dists.values(), start_e_dists.values())])
+        if not (all_pmfs or all_pdfs):
+            raise ValueError('e_dists and start_e_dists are not all rvs of the same type.')
 
         # Check start_dists
         start_states = set(start_t_dist)
@@ -347,7 +375,7 @@ class ARHMM:
         # Create random variates from t_dists
         state2idx = {}
         idx2state = {}
-        for idx, state in enumerate(states):
+        for idx, state in enumerate(sorted(states)):
             state2idx[state] = idx
             idx2state[idx] = state
         t_dists_rv = {}
@@ -362,6 +390,11 @@ class ARHMM:
         start_t_dist_rv = rv_from_dict(start_t_dist, state2idx)
         start_e_dists_rv = {state2idx[state]: e_dist for state, e_dist in start_e_dists.items()}
 
+        # Extract probability functions from e_dists
+        attr = 'pdf' if all_pdfs else 'pmf'
+        e_dists_pf = {state: getattr(e_dist, attr) for state, e_dist in e_dists_rv.items()}
+        start_e_dists_pf = {state: getattr(e_dist, attr) for state, e_dist in start_e_dists_rv.items()}
+
         self.name = name
         self.states = states
         self._states = set(idx2state)
@@ -371,17 +404,19 @@ class ARHMM:
         self._t_dists_rv = t_dists_rv
         self.e_dists = e_dists
         self._e_dists_rv = e_dists_rv
+        self._e_dists_pf = e_dists_pf
         self.start_t_dist = start_t_dist
         self._start_t_dist_rv = start_t_dist_rv
         self.start_e_dists = start_e_dists
         self._start_e_dists_rv = start_e_dists_rv
+        self._start_e_dists_pf = start_e_dists_pf
         self.start_states = start_states
         self._start_states = {state2idx[state] for state in self.start_states}
         self.stop_states = stop_states
         self._stop_states = [state2idx[state] for state in stop_states] if stop_states is not None else []
 
     def __repr__(self):
-        pad = 4 * ' '
+        pad = 6 * ' '
         return (f"ARHMM(states={self.states},\n"
                 f"{pad}stop_states={self.stop_states},\n"
                 f"{pad}name='{self.name}')")
@@ -389,34 +424,34 @@ class ARHMM:
     def simulate(self, step_max):
         if step_max == 0:
             return []
-        s0 = self._start_t_dist_rv.rvs()
-        e0 = self._start_e_dists_rv[s0].rvs()
-        steps = [(self._idx2state[s0], e0)]
+        state0 = self._start_t_dist_rv.rvs()
+        emit0 = self._start_e_dists_rv[state0].rvs()
+        steps = [(self._idx2state[state0], emit0)]
         for i in range(step_max-1):
-            if s0 in self._stop_states:
+            if state0 in self._stop_states:
                 return steps
-            s1 = self._t_dists_rv[s0].rvs()
-            e1 = self._e_dists_rv[s1].rvs(e0)
-            steps.append((self._idx2state[s1], e1))
-            s0 = s1
-            e0 = e1
+            state1 = self._t_dists_rv[state0].rvs()
+            emit1 = self._e_dists_rv[state1].rvs(emit0)
+            steps.append((self._idx2state[state1], emit1))
+            state0 = state1
+            emit0 = emit1
         return steps
 
     def viterbi(self, emits):
         # Forward pass
         vs = {state: [(0, [None])] for state in self._states}
         for state in self._start_states:
-            vs[state] = [(log(self._start_e_dists_rv[state].pmf(emits[0])) + log(self._start_t_dist_rv.pmf(state)), [None])]
+            vs[state] = [(log(self._start_e_dists_pf[state](emits[0])) + log(self._start_t_dist_rv.pmf(state)), [None])]
         for i, emit in enumerate(emits[1:]):
-            for state in self._states:
+            for state0 in self._states:
                 # Get probabilities
-                t_probs = {s: vs[s][i][0] + log(self._t_dists_rv[s].pmf(state)) for s in self._states}
+                t_probs = {state1: vs[state1][i][0] + log(self._t_dists_rv[state1].pmf(state0)) for state1 in self._states}
                 t_prob = max(t_probs.values())  # Probability of most likely path to state
-                e_prob = log(self._e_dists_rv[state].pmf(emits[i], emit))
+                e_prob = log(self._e_dists_pf[state0]((emits[i], emit)))
 
                 # Get traceback states
                 tb_states = [s for s, p in t_probs.items() if p == t_prob]
-                vs[state].append((e_prob+t_prob, tb_states))
+                vs[state0].append((e_prob+t_prob, tb_states))
 
         # Compile traceback states (taking care to allow for multiple paths)
         v_max = max([v[-1][0] for v in vs.values()])
@@ -438,7 +473,7 @@ class ARHMM:
         # Initialize
         fs = {state: [0] for state in self._states}
         for state in self._start_states:
-            fs[state] = [self._start_e_dists_rv[state].pmf(emits[0]) * self._start_t_dist_rv.pmf(state)]
+            fs[state] = [self._start_e_dists_pf[state](emits[0]) * self._start_t_dist_rv.pmf(state)]
         s = sum([fs[state][0] for state in self._states])
         for state in self._states:
             fs[state][0] /= s
@@ -447,11 +482,11 @@ class ARHMM:
         # Forward pass
         for i, emit in enumerate(emits[1:]):
             # Get probabilities
-            for state in self._states:
-                t_probs = [fs[s][i] * self._t_dists_rv[s].pmf(state) for s in self._states]
+            for state0 in self._states:
+                t_probs = [fs[state1][i] * self._t_dists_rv[state1].pmf(state0) for state1 in self._states]
                 t_prob = sum(t_probs)  # Probability of all paths to state
-                e_prob = self._e_dists_rv[state].pmf(emits[i], emit)
-                fs[state].append(e_prob*t_prob)
+                e_prob = self._e_dists_pf[state0]((emits[i], emit))
+                fs[state0].append(e_prob*t_prob)
 
             # Scale probabilities
             s = sum([fs[state][i+1] for state in self._states])
@@ -475,10 +510,10 @@ class ARHMM:
         # Backward pass
         for i, emit in enumerate(emits[:0:-1]):  # Reverse sequence starting from last emit excluding first
             # Get probabilities
-            for state in self._states:
-                probs = [bs[s][i] * self._t_dists_rv[state].pmf(s) * self._e_dists_rv[s].pmf(emits[-(i+2)], emit) for s in self._states]
+            for state0 in self._states:
+                probs = [bs[state1][i] * self._t_dists_rv[state0].pmf(state1) * self._e_dists_pf[state1]((emits[-(i+2)], emit)) for state1 in self._states]
                 prob = sum(probs)  # Probability of all paths to state
-                bs[state].append(prob)
+                bs[state0].append(prob)
 
             # Scale probabilities
             s = sum([bs[state][i+1] for state in self._states])
